/******************************************************************************/
/*  Virtual File System ZIP                                                   */
/*     -   Fredrik Redgrd <mrfrippe@writeme.com>                             */
/******************************************************************************/

#include <windows.h>
#include <string.h>

#include "filesystem.h"
#include "unzip.h"

/////////////////////////////////////////////////////////////////////////////////
//defines///////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////
//local function prototypes//////////////////////////////////////////////////////

int AddFoundFile( const char *path, LPWIN32_FIND_DATA info );
int FindFile( const char *filename );
unsigned long HashString(const char *string);

/////////////////////////////////////////////////////////////////////////////////
//local types////////////////////////////////////////////////////////////////////

typedef struct 
{
	char *filename;
	unzFile file;
} ZipFile;

typedef struct
{
	char filename[256];
	ZipFile *container;

	unsigned long hash;
	unsigned long filesize;
	int num;
	int pos;

	unsigned long dos_date;

} FileItem;

/////////////////////////////////////////////////////////////////////////////////
//global variables///////////////////////////////////////////////////////////////

char *loose_files_path;
unsigned int num_zipfiles;
unsigned int num_files;

ZipFile *zipfiles;
FileItem *files;

/////////////////////////////////////////////////////////////////////////////////
//functions//////////////////////////////////////////////////////////////////////

unsigned long HashString(const char *string)
{
	unsigned long result = 0;
	int place = 0;
	while (*string)
	{
		char string_value = isupper(*string) ? tolower(*string) : *string;
		result += (string_value << (place * 8));
		string++;
		place++;
		if (place > 3) place = 0;
	}
	return result;
}

int FindFile( const char *filename )
{
	unsigned int i;
	unsigned long hash = HashString(filename);
	
	unsigned long hash_l = HashString("file.txt");
	unsigned long hash_u = HashString("File.txt");


	for (i=0; i<num_files; i++)
	{
		if (files[i].hash == hash)
		{
			if (!_stricmp(files[i].filename, filename))
			{
				return i;
			}
		}
	}

	return -1;
}

unsigned long vfsGetFileSize( const char *filename )
{
	if (loose_files_path)
	{
		//first look for loose files
		char *loose_name;
		int len = strlen(loose_files_path) + strlen(filename)+1;
		loose_name = (char *)malloc(len);
		strcpy(loose_name, loose_files_path);
		strcpy(loose_name+strlen(loose_name),filename);

		HANDLE lf;
		lf = CreateFile(loose_name, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		if ((lf==0) || (lf==INVALID_HANDLE_VALUE))
		{
		}
		else
		{
			unsigned long ret = GetFileSize(lf, NULL);
			CloseHandle(lf);
			free(loose_name);
			return ret;
		}
		free(loose_name);
	}

	int find = FindFile(filename);
	if (find == -1) return 0;

	return files[find].filesize;
}

unsigned long vfsLoadFileIntoBuffer( const char *filename, void *buffer, unsigned long buffer_size )
{
	//NOTE: the buffer should allready be allocated. use vfsGetFileSize to get required size.

	if (loose_files_path)
	{ //loose files
		char *loose_name;
		int len = strlen(loose_files_path) + strlen(filename)+1;
		loose_name = (char *)malloc(len);
		strcpy(loose_name, loose_files_path);
		strcpy(loose_name+strlen(loose_name),filename);

		HANDLE lf;
		lf = CreateFile(loose_name, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		if ((lf==0) || (lf==INVALID_HANDLE_VALUE))
		{
		}
		else
		{
			unsigned long lf_read;
			unsigned long ret = GetFileSize(lf, NULL);
			ReadFile(lf, buffer, ret, &lf_read, NULL);
			CloseHandle(lf);
			free(loose_name);
			return lf_read;
		}
		free(loose_name);
	}

	int find = FindFile(filename);
	if (find == -1) return 0;

	unzLocateFileMy( files[find].container->file, files[find].num, files[find].pos);
	unzOpenCurrentFile( files[find].container->file);
	unsigned long read = unzReadCurrentFile( files[find].container->file, buffer, buffer_size);
	unzCloseCurrentFile( files[find].container->file );

	return read;
}

int vfsBootup()
{
	loose_files_path = 0;
	num_zipfiles = 0;
	num_files = 0;

	files = 0;
	zipfiles = 0;

	return 1;
}

void vfsShutdown()
{
	unsigned int i;

	for (i=0; i<num_zipfiles; i++)
	{
		unzClose(zipfiles[i].file);
//		free(zipfiles[i].filename); why can't I free this
	}

	free(files);		
}

int vfsAddPackfileSearchDir( const char *path )
{
	HANDLE file_search;
	WIN32_FIND_DATA fd;
	int found_files = 0;

	char *buffer = new char[strlen(path)+10];

	memcpy(buffer, path, strlen(path)+1);
	strcat(buffer, "*.zip");

	file_search = FindFirstFile( buffer, &fd );
	if (file_search == INVALID_HANDLE_VALUE)
	{
		DWORD error = GetLastError();
		return 0;
	}
	found_files+=AddFoundFile( path, &fd );
	

	while (FindNextFile( file_search, &fd ))
	{
		found_files+=AddFoundFile( path, &fd );
	}
	FindClose(file_search);

	return found_files;
}

int AddFoundFile( const char *path, LPWIN32_FIND_DATA info )
{
	unsigned int i,j;

	////// get file name //////////
	char *zipfilename;
	int zfn_len;
	zfn_len = strlen(path) + strlen(info->cFileName);

	zipfilename = (char*)malloc( zfn_len );
	
	strcpy(zipfilename, path);
	strcpy(zipfilename+strlen(path), info->cFileName);
	

	////// Add zip to list //////

	for (i=0; i< num_zipfiles; i++)
	{
		if (!_stricmp( zipfiles[i].filename, zipfilename)!=0)
			return 0;
	}

	////// open it up ///////////
	unzFile f = unzOpen(zipfilename);
	if (f==0)
		return 0;


	if (zipfiles == 0)
		zipfiles = (ZipFile*)malloc(sizeof(ZipFile));
	else
		zipfiles = (ZipFile*)realloc(zipfiles,sizeof(ZipFile)*(num_zipfiles+1));

	zipfiles[num_zipfiles].filename = zipfilename;
	zipfiles[num_zipfiles].file = f;

	num_zipfiles++;


	/////// Get File info ///////

	unz_global_info ugi;
	unzGetGlobalInfo(f, &ugi);

	if (files == 0)
		files = (FileItem*)malloc(sizeof(FileItem) * ugi.number_entry);
	else
		files = (FileItem*)realloc(files, sizeof(FileItem) * (num_files + ugi.number_entry));
	//it doesn't matter if there are file extra spaces allocated, since the next realloc would
	//just make enough free file spaces to fit in the new files

	for (i=0; i<ugi.number_entry; i++)
	{
		char *filename = (char*)malloc(256);
		unz_file_info ufi;
		unzGetCurrentFileInfo(f, &ufi, filename, 255, NULL, 0, NULL, 0);

		for (j=0; j<num_files; j++)
		{
			if (!_stricmp(filename, files[j].filename)) 
			{
				if (ufi.dosDate > files[j].dos_date) //newer file, replace
				{
					//strcpy(files[j].filename, filename);
					files[j].num = i;
					files[j].pos = ufi.c_offset;
					files[j].dos_date = ufi.dosDate;
					files[j].filesize = ufi.uncompressed_size;
					files[j].container = &zipfiles[num_zipfiles-1];
					files[j].hash = HashString(filename);
				}
				goto next_file;
			}
		}
		
		// new file, add it
		strcpy(files[num_files].filename, filename);
		files[num_files].num = i;
		files[num_files].pos = ufi.c_offset;
		files[num_files].dos_date = ufi.dosDate;
		files[num_files].filesize = ufi.uncompressed_size;
		files[num_files].container = &zipfiles[num_zipfiles-1];
		files[j].hash = HashString(filename);
		num_files++;

next_file:
		free(filename);
		unzGoToNextFile(f);
	}

	return 1;
}

int vfsSetLooseFilesBaseDir( const char *path )
{
	int len = strlen(path);

	if (loose_files_path)
		free(loose_files_path);

	loose_files_path = (char *)malloc(len+1);
	strcpy(loose_files_path, path);
	loose_files_path[len] = 0;

	return 1;
}